/**
 * A helper returning constructor class of given argument.
 */
export const $classOf = (/** @type {any} */ arg) => arg?.constructor

/**
 * A helper that checks if the given argument is of type `string`.
 *
 * @returns {arg is string}
 */
export const $isString = (/** @type {any} */ arg) => $classOf(arg) === String

/**
 * A helper that checks if the given argument is of type `number`.
 *
 * @returns {arg is number}
 */
export const $isNumber = (/** @type {any} */ arg) => $classOf(arg) === Number

/**
 * A helper that checks if the given argument is of type `Record<string, any>`.
 *
 * @returns {arg is Record<string, any>}
 */
export const $isObject = (/** @type {any} */ arg) => $classOf(arg) === Object

/**
 * A helper that checks if the given argument is of type `Array`.
 */
export const $isArray = Array.isArray

/**
 * Replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.
 */
export const $in = (/** @type {any} */ key, /** @type {any} */ ref) =>
  ($isString(key) || $isNumber(key)) && Object.hasOwnProperty.call(ref ?? Object, key)

/**
 * A helper that checks the equality of the given arguments.
 *
 * @returns {boolean}
 */
export const $eq = (/** @type {any} */ x, /** @type {any} */ y) => {
  if (x === y) {
    return true
  }

  const classOfX = $classOf(x)

  if (classOfX === $classOf(y)) {
    if (classOfX === Number) {
      return (isNaN(x) && isNaN(y)) || +x === +y
    }

    if (classOfX === Date) {
      return +x === +y
    }

    if (classOfX === String || classOfX === RegExp) {
      return '' + x === '' + y
    }

    if (classOfX === Array) {
      return x.length === y.length && x.every((/** @type {any} */ item, /** @type {number} */ index) => $eq(item, y[index]))
    }

    if (classOfX === Object) {
      const keysOfX = Object.keys(x)
      return keysOfX.length === Object.keys(y).length && keysOfX.every(key => $eq(x[key], y[key]))
    }

    if (classOfX === Set || classOfX === Map) {
      if (x.size !== y.size) {
        return false
      }
      const xa = [...x]
      const ya = [...y]
      return xa.every(xv => ya.find(yv => $eq(xv, yv)))
    }
  }

  return false
}

/**
 * @typedef {keyof HTMLElementTagNameMap | HTMLElement | ShadowRoot} $HElement
 */

/**
 * @typedef {Record<string, any>} $Keys
 */

/**
 * @typedef {$HElement | Text | SVGElement | string | number | $HConfig} $HItem
 */

/**
 * @typedef {[$HElement, $Keys, ...$HItem[]]} $HConfig
 */

/**
 * A lightweight ExtJS-like helper for creating and modifying `HTMLElement`s.
 *
 * - `e`: the modified `HTMLElement` or tag of the element being created.
 * - `keys`: map of properties and/or attributes. Keys starting with `$` are treated as *properties* to be set on the element being created or modified (without the leading `$`). All other keys are treated as *attributes* to be set by `setAttribute`. Attributes equal to `false` cause the attribute to be removed by `removeAttribute`.
 * - `items`: items to be appended to the element being created or modified.
 *
 * @type {{
 *   <T extends keyof HTMLElementTagNameMap>(e: T, keys?: $Keys, ...items: $HItem[]): HTMLElementTagNameMap[T];
 *   <E extends HTMLElement>(e: E, keys?: $Keys, ...items: $HItem[]): E;
 *   (e: $HElement, keys?: $Keys, ...items: $HItem[]): HTMLElement | ShadowRoot;
 * }}
 *
 * @returns {HTMLElement | ShadowRoot}
 */
export const $h = (/** @type {$HElement} */ e, /** @type {$Keys} */ keys = {}, /** @type {$HItem[]} */...items) => {
  e = $isString(e) ? document.createElement(e) : e

  if (e instanceof HTMLElement) {
    for (const key in keys) {
      const value = keys[key]
      if (key[0] === '$') {
        const key1 = key.slice(1)
        if (key1 in e && $isObject(value)) { // warning: $in cannot be used here!
          for (const key2 in value) {
          // @ts-expect-error:
            e[key1][key2] = value[key2]
          }
        } else {
        // @ts-expect-error:
          e[key1] = value
        }
      } else {
        if (value === true) {
          e.setAttribute(key, '')
        } else if (value === false) {
          e.removeAttribute(key)
        } else {
          e.setAttribute(key, value)
        }
      }
    }
  }

  for (const item of items) {
    e.appendChild(item instanceof HTMLElement || item instanceof SVGElement || item instanceof Text || item instanceof ShadowRoot
      ? item
      : $isString(item) || $isNumber(item)
        // @ts-expect-error:
        ? document.createTextNode(item)
        : $h(...item))
  }

  return e
}

/**
 * @typedef {keyof SVGElementTagNameMap | SVGElement} $SElement
 */

/**
 * @typedef {$SElement | Text | string | number | $SConfig} $SItem
 */

/**
 * @typedef {[$SElement, $Keys, ...$SItem[]]} $SConfig
 */

const /** @type {Record<string, string>} */ NS = {
  xlink: 'http://www.w3.org/1999/xlink'
}

/**
 * A lightweight ExtJS-like helper for creating and modifying `SVGElement`s (conf. `$h`).
 *
 * - `e`: the modified `SVGElement` or tag of the element being created.
 * - `keys`: map of properties and/or attributes. Keys starting with `$` are treated as *properties* to be set on the element being created or modified (without the leading `$`). All other keys are treated as *attributes* to be set by `setAttribute` or `setAttributeNS`. Attributes equal to `false` cause the attribute to be removed by `removeAttribute` or `removeAttributeNS`.
 * - `items`: items to be appended to the element being created or modified.
 *
 * @type {{
 *   <T extends keyof SVGElementTagNameMap>(e: T, keys?: $Keys, ...items: $SItem[]): SVGElementTagNameMap[T];
 *   <E extends SVGElement>(e: E, keys?: $Keys, ...items: $SItem[]): E;
 *   (e: $SElement, keys?: $Keys, ...items: $SItem[]): SVGElement;
 * }}
 *
 * @returns {SVGElement}
 */
export const $s = (/** @type {$SElement} */ e, /** @type {$Keys} */ keys = {}, /** @type {$SItem[]} */...items) => {
  e = $isString(e) ? document.createElementNS('http://www.w3.org/2000/svg', e) : e

  for (const key in keys) {
    const value = keys[key]

    if (key[0] === '$') {
      const key1 = key.slice(1)
      if (key1 in e && $isObject(value)) { // warning: $in cannot be used here!
        for (const key2 in value) {
          // @ts-expect-error:
          e[key1][key2] = value[key2]
        }
      } else {
        // @ts-expect-error:
        e[key1] = value
      }
    } else {
      const indexOfColon = key.indexOf(':')

      if (indexOfColon >= 0) {
        const /** @type {string=} */ ns = NS[key.slice(0, indexOfColon)]

        if (ns != null) {
          const name = key.slice(indexOfColon + 1)

          if (value === true) {
            e.setAttributeNS(ns, name, '')
          } else if (value === false) {
            e.removeAttributeNS(ns, name)
          } else {
            e.setAttributeNS(ns, name, $isString(value) ? value : '' + value)
          }
        }
      } else {
        if (value === true) {
          e.setAttribute(key, '')
        } else if (value === false) {
          e.removeAttribute(key)
        } else {
          e.setAttribute(key, $isString(value) ? value : '' + value)
        }
      }
    }
  }

  for (const item of items) {
    e.appendChild(item instanceof SVGElement || item instanceof Text
      ? item
      : $isString(item) || $isNumber(item)
        // @ts-expect-error:
        ? document.createTextNode(item)
        : $s(...item))
  }

  return e
}

/**
 * @typedef {{
 *   [attributeOrSelector: string]: string | number | $CSSRules;
 * }} $CSSRules
 */

/**
 * @typedef {{
 *   [selector: string]: $CSSRules;
 * }} $CSS
 */

const _c = (
  /** @type {$CSSRules} */ style,
  /** @type {string} */ prefix,
  /** @type {string[]} */ result,
  /** @type {(text: string) => string} */ split
) => {
  const /** @type {[$CSSRules | string[], string][]} */ queue = [[style, prefix]]

  while (queue.length > 0) {
    const [style2, prefix2] = queue.shift() ?? []

    if (style2 == null || prefix2 == null) {
      continue
    }

    if ($isArray(style2)) {
      result.push(prefix2, prefix2 !== '' ? '{' : '', style2.join(';'), prefix2 !== '' ? '}' : '')
    } else {
      const /** @type {[$CSSRules | string[], string][]} */ todo = []
      let /** @type {string[]} */ attributes = []
      let attributesPushed = false

      for (const key in style2) {
        const value = style2[key]

        if ($isString(value) || $isNumber(value)) {
          if (!attributesPushed) {
            attributesPushed = true
            attributes = []
            todo.push([attributes, prefix2])
          }

          attributes.push(`${split(key).replace(/([A-Z])/g, (_, letter) => '-' + letter.toLowerCase())}:${value}`)
        } else {
          attributesPushed = false

          const /** @type {string[]} */ newPrefix = []
          const keySplitted = key.split(',')

          for (const prefixItem of prefix2.split(',')) {
            for (const keyItem of keySplitted) {
              newPrefix.push(prefixItem + keyItem)
            }
          }

          todo.push([value, newPrefix.join(',')])
        }
      }

      queue.unshift(...todo)
    }
  }
}

/**
 * CSS-in-JS helper based on the `$CSS` format.
 * An iterative implementation that does not cause a stack overflow exception.
 *
 * The `style` parameter provides a hierarchical description of CSS rules.
 *
 * - Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
 * - All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
 * - In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
 * - Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
 * - Top-level keys that begin with `@` are not concatenated with sub-object keys.
 */
export const $c = (/** @type {$CSS} */ style, /** @type {string} */ splitter = '$$') => {
  const split = (/** @type {string} */ text) => text.split(splitter)[0]
  const /** @type {string[]} */ result = []

  for (const key in style) {
    const value = style[key]

    if (key[0] === '@') {
      result.push(split(key) + '{')
      _c(value, '', result, split)
      result.push('}')
    } else {
      _c(value, split(key), result, split)
    }
  }

  return result.join('')
}

/**
 * A generic helper for syntax highlighting.
 *
 * @returns {$HItem[]}
 */
export const $highlight = (
  /** @type {RegExp} */ pattern,
  /** @type {((chunk: string, index: number) => $HItem)[]} */ highlighters,
  /** @type {string} */ innerText
) => {
  const /** @type {$HItem[]} */ result = []

  innerText.split(pattern).forEach((chunk, index) => {
    index %= highlighters.length
    if (chunk != null) {
      result.push(highlighters[index](chunk, index))
    }
  })

  return result
}

/** @returns {string} */
const _locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ language,
  /** @type {string} */ text,
  /** @type {string} */ version
) => {
  // @ts-expect-error:
  const v = locales?.[language]?.[version]?.[text]
  if ($isString(v)) {
    return v
  }
  const t = locales?.[language]?.[text]
  return $isString(t) ? t : text
}

/**
 * Language translations helper.
 */
export const $locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ defaultLanguage,
  /** @type {readonly string[]} */ languages = navigator.languages
) => {
  for (const language of languages) {
    if ($in(language, locales)) {
      // @ts-expect-error:
      return _locale.bind(0, locales, language)
    }
  }

  // @ts-expect-error:
  return _locale.bind(0, locales, defaultLanguage)
}
